/* Copyright (c) 2010, NICTA 
 * All rights reserved. 
 * 
 * The contents of this file are subject to the Mozilla Public License 
 * Version 1.1 (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at 
 * http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the 
 * License for the specific language governing rights and limitations 
 * under the License. 
 * 
 * Author: James Petterson (james.petterson@nicta.com.au)
 */

#include "common.hpp"
#include "sml.hpp"
#include "configuration.hpp"
#include "bmrmexception.hpp"
#include "timer.hpp"
#include "genericdata.hpp"
#include <vector>
#include <math.h>
#include "preader.h"
#include "words.h"
#include "mt.h"

using namespace std;

CGenericData::CGenericData(unsigned int start, unsigned int nparts)
{
	Configuration &config    = Configuration::GetInstance();
	bool   training          = config.GetBool("Training");
	string filename_x        = config.GetString("Data.xFile");
	string filename_y        = config.GetString("Data.yFile");
	char  *line;
	bool   const_feature     = true; /* constant feature to have a bias term */
	int    D                 = config.IsSet("random_features") ? config.GetInt("random_features") : 0;
	int    sparse_features   = config.IsSet("sparse_features") ? config.GetInt("sparse_features") : 0;
	int    x_dim_linear;
	double *F=NULL, *new_descriptor=NULL;

	printf("Reading features..."); fflush(stdout);

	/* read features */
	preader r1(filename_x.c_str());
	while (1)
	{
		line = r1.get_line();
		if (!line) break;

		words w(line, false);

		/* first word is filename; second is 1 for training sample, 0 for testing sample; third and on are features */
		if (w.count()<3) break;

		/* get only training (testing) samples */
		if (training!=atoi(w.begin(1))) continue;

		if (sparse_features)
		{
			/* note: the code assumes that feature indexes are in the range 1 to sparse_features-1 */
			int n_features = w.count()-2;
			if (const_feature) n_features++;

			double *f_vals = new double[n_features+1];
			int    *f_idxs = new int[n_features+1];
			int     o_pos  = 0;
			int     i_pos  = 2;

			if (const_feature)
			{
				f_vals[o_pos] = 1.0;
				f_idxs[o_pos] = 0;
				o_pos++;
			}

			for (; o_pos<n_features; o_pos++, i_pos++)
			{
				char *idx_s = w.begin(i_pos);
				char *val_s = strchr(idx_s, ':');
				int   idx   = atoi(idx_s);

				if (val_s==NULL)
				{
					printf("invalid entry: %s\n", w.begin(i_pos));
					exit(-1);
				}
				if ((idx<1)||(idx>=sparse_features))
				{
					printf("feature index out of range: expecting 1-%d, got %d\n", sparse_features-1, idx);
					exit(-1);
				}
				
				f_vals[o_pos] = atof(++val_s);
				f_idxs[o_pos] = idx;
			}

			features_idxs.push_back(f_idxs);
			features_vals.push_back(f_vals);
			features_nnz.push_back(n_features);
			x_dim = sparse_features;
		}
		else
		{
			/* assumes all examples have the same dimensionality */
			x_dim_linear = w.count()-2;

			if (const_feature)
				x_dim_linear++;

			x_dim     = x_dim_linear+2*D;
			double *f = new double[x_dim];

			if (const_feature)
			{
				f[0] = 1.0;
				for (int i=0; i<x_dim_linear-1; i++)
					f[i+1] = atof(w.begin(2+i));
			}
			else
			{
				for (int i=0; i<x_dim; i++)
					f[i] = atof(w.begin(2+i));
			}

			if (D>0)
			{
				if (F==NULL)
				{
					/* matrix with samples from a standard gaussian distribution N(0,1) using Box-Muller transform */
					F = new double[D*x_dim_linear];
					new_descriptor = new double[D];

					seedMT(0);
					for (int i=0; i<D*x_dim_linear; i++)
						F[i] = 	sqrt(-2*log(frandomMT()))*cos(2*M_PI*frandomMT());
				}

				/* multiply original features by F */
				for (int i=0; i<D; i++)
				{
					double sum = 0;
					for (int j=0; j<x_dim_linear; j++)
						sum += f[j]*F[i*x_dim_linear+j];
					new_descriptor[i] = sum;
				}

				/* add the new features to f */
				for (int i=0; i<D; i++)
				{
					f[x_dim_linear+i]   = cos(new_descriptor[i]);
					f[x_dim_linear+i+D] = sin(new_descriptor[i]);
				}
			}
			features.push_back(f);
		}
	}

	if (sparse_features && D)
	{
		/* sparse_features with kernel approximation using random features:
		 * we will create a dense vector for the random features and discard the sparse ones */

		if (F==NULL)
		{
			/* matrix with samples from a standard gaussian distribution N(0,1) using Box-Muller transform */
			F = new double[D*x_dim];
			seedMT(0);
			for (int i=0; i<D*x_dim; i++)
				F[i] = 	sqrt(-2*log(frandomMT()))*cos(2*M_PI*frandomMT());
		}

		for (int n=0; n<(int)features_nnz.size(); n++)
		{
			double *vals = features_vals[n];
			int    *idxs = features_idxs[n];
			int     nnz  = features_nnz[n];
			double *f    = new double[2*D];

			/* multiply original features by F */
			for (int i=0; i<D; i++)
			{
				double sum = 0;
				for (int j=0; j<nnz; j++)
					sum += vals[j]*F[i*x_dim+idxs[j]];
				f[i]   = cos(sum);
				f[i+D] = sin(sum);
			}

			features.push_back(f);
		}

		sparse_features = 0; /* signal to the rest of the code that we have only dense features */
		x_dim = 2*D;
	}

	if (sparse_features)
		N = features_nnz.size();
	else
		N = features.size();

	/* read labels */
	vocab_size = 0;
	preader r2(filename_y.c_str());
	while (1)
	{
		line = r2.get_line();
		if (!line) break;

		words w(line, false);

		/* first word is filename; second is 1 for training sample, 0 for testing sample; third and on are tags */
		if (w.count()<2) break;

		/* get only training (testing) samples, but use all samples to compute the vocabulary size */
		if (training!=atoi(w.begin(1)))
		{
			int y_dim = w.count()-2;
			int tag;
			for (int i=0; i<y_dim; i++)
			{
				tag = atoi(w.begin(2+i));
				if (tag>vocab_size) vocab_size = tag;
			}

		}
		else
		{
			int y_dim = w.count()-2;
			int    *t = new int[y_dim+1];
			t[0]      = y_dim; /* first position is tag count, second and on are tag values */
			for (int i=0; i<y_dim; i++)
			{
				t[i+1] = atoi(w.begin(2+i));
				if (t[i+1]>vocab_size) vocab_size = t[i+1];
			}

			tags.push_back(t);
		}
	}
	vocab_size++;

	if ((int)tags.size()!=N)
	{
		printf("size(x) = %d, size(y)=%ld !\n", N, tags.size());
		exit(-1);
	}

	printf("done\nN = %d, dim = %d, vocab_size = %d\n", N, x_dim, vocab_size);

	/* for each tag, list the images that contain it */
	images_with_tag = new int*[vocab_size];
	for (int tag=0; tag<vocab_size; tag++)
	{
		images_with_tag[tag]    = new int[N+1];
		images_with_tag[tag][0] = 0;
	}

	for (int image_nr=0; image_nr<N; image_nr++)
	{
		for (int i=0; i<tags[image_nr][0]; i++)
		{
			int tag    = tags[image_nr][i+1];
			int *count = &images_with_tag[tag][0];
			images_with_tag[tag][++(*count)] = image_nr;
		}
	}
}

CGenericData::~CGenericData()
{
}

