/* Copyright (c) 2010, NICTA 
 * All rights reserved. 
 * 
 * The contents of this file are subject to the Mozilla Public License 
 * Version 1.1 (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at 
 * http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the 
 * License for the specific language governing rights and limitations 
 * under the License. 
 * 
 * Author: James Petterson (james.petterson@nicta.com.au)
 */

#include "genericdata.hpp"
#include "genericloss.hpp"
#include "configuration.hpp"
#include "math.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>
#ifdef USE_OPENMP
#include <omp.h>
#endif
#ifdef USE_MKL
#include <mkl_cblas.h>
#endif
#include "mt.h"

/** Constructor. */
	CGenericLoss::CGenericLoss(int dim, CModel* &model_, CGenericData* &data_)
: CLoss(model_, 1, dim, data_->bias()), data(data_)
{
	srand(0);
	Configuration &config = Configuration::GetInstance();

	verbosity = config.GetInt("Loss.verbosity");
	sparse_features = config.IsSet("sparse_features") ? config.GetInt("sparse_features") : 0;

	if (config.IsSet("random_features"))
		sparse_features = 0; /* when we use both sparse features and random features we get a dense vector */

	if (config.IsSet("max_sparsity"))
		max_sparsity = config.GetInt("max_sparsity");
	else
		max_sparsity = -1;

	if (config.IsSet("score_to_optimize"))
		score_to_optimize = config.GetInt("score_to_optimize");
	else
		score_to_optimize = MACRO_FBETA;

	if (config.IsSet("beta"))
		beta_sq = pow(config.GetDouble("beta"),2);
	else
		beta_sq = 1.0; /* macro-F1 */


	if (score_to_optimize==HAMMING_LOSS)
		max_sparsity = -1; /* to make sure it is set to N, since this variable is used to allocate y/ybar */

	if ((max_sparsity>data->N)||(max_sparsity<0)) max_sparsity = data->N;
}

/** Destructor. */
CGenericLoss::~CGenericLoss() 
{
}


static int compare_scores(const void *p1, const void *p2)
{
	double diff = ((score_t *)p1)->score - ((score_t *)p2)->score;
	return (diff==0) ? 0 : ((diff>0) ? -1 : +1);
}

static int compare_int(const void *p1, const void *p2)
{
	return *((int *)p1) - *((int *)p2);
}


/* loss for correct label y and predicted label ybar */
double CGenericLoss::label_loss(int *y, int *ybar, double *precision, double *recall)
{
	int sparsity_y    = y[0];
	int sparsity_ybar = ybar[0];
	int i1, i2;
	double loss=0;

	if (score_to_optimize==MACRO_FBETA)
	{
		/* compute the number of matches (in the precision/recall sense - i.e., the dot product of y and ybar);
		 * since they are stored in sparse format we need the code below;
		 * note: we assume both vectors are sorted */
		int num_matches = 0;
		i1 = i2 = 1;
		while ((i1 <= sparsity_y) && (i2 <= sparsity_ybar))
		{
			if (y[i1]==ybar[i2])
			{
				num_matches++;
				i1++;
				i2++;
			}
			else if (y[i1]<ybar[i2])
				i1++;
			else
				i2++;
		}

		/* our loss is the f-score:
		 *
		 * loss(y, ybar) = 1 - (1+beta^2) (y^T ybar) / (ybar^T ybar + beta^2 y^T y)
		 */
		loss = 1.0-((1+beta_sq)*(double)num_matches/((double)sparsity_ybar+(double)sparsity_y*beta_sq));

		if (precision)
			*precision += (sparsity_ybar>0) ? (double)num_matches/(double)sparsity_ybar : 0.0;

		if (recall)
			*recall += (sparsity_y>0) ? (double)num_matches/(double)sparsity_y : 0.0;
	}

	if (score_to_optimize==HAMMING_LOSS)
	{
		/* compute the number of mismatches mismaches (y_i != ybar_i);
		 * since they are stored in sparse format we need the code below;
		 * note: we assume both vectors are sorted */
		i1 = i2 = 1;
		int num_mismatches = 0;

		int Z = data->N;

		while ((i1 <= sparsity_y) && (i2 <= sparsity_ybar))
		{
			if (y[i1]==ybar[i2])
			{
				i1++;
				i2++;
			}
			else if (y[i1]<ybar[i2])
			{
				i1++;
				num_mismatches++;
			}
			else
			{
				i2++;
				num_mismatches++;
			}
		}
		num_mismatches += (sparsity_y-i1) + (sparsity_ybar-i2) + 2;

		loss = (double)num_mismatches/(double)Z;
	}

	return loss;
}

/* 
 * Check if 'sparsity' is a lower bound on ybar's sparsity.
 * Return:
 *   >0 if yes (number of positives entries in z_n)
 *    0 otherwise
 */
int CGenericLoss::is_sparsity_bound(int sparsity, int gt_sparsity, double *image_scores, int *dense_y, double beta_sq)
{
	int num_pos = 0;
	double score, const_factor = -(1+beta_sq)/(beta_sq*gt_sparsity+sparsity);

	for(int image_nr=0; image_nr<data->N; image_nr++)
	{
		score = image_scores[image_nr];
		if (dense_y[image_nr]==1) score += const_factor;
		if (score>0) num_pos++;
	}

	return (num_pos>sparsity) ? num_pos : 0;
}


/* 
 * If only_classify=false, find the y that generates the most violated constraint for tag 'tag' and save it in ybar:
 *
 * ybar = argmax_y loss(y, y^n) + <phi(x^n, y), theta>
 *
 *
 * If only_classify=true, find the most likely y for tag 'tag' and save it in ybar:
 *
 * ybar = argmax_y <phi(x^n, y), theta>
 *
 */
void CGenericLoss::find_most_violated_constraint(int tag, int *ybar, TheMatrix &w, bool only_classify)
{
	int     *y    = data->images_with_tag[tag];
	double *theta = w.Data();
			

	score_t scores[data->N]; /* score for each image */
	double  image_scores[data->N]; /* score for each image (just the dot product, can be reused) */
	int         gt_sparsity = y[0];  /* number of images in the ground truth */

	/* scores for each image */
	memset(scores,       0, sizeof(scores));
	memset(image_scores, 0, sizeof(image_scores));

	/* convert ground-truth to dense format for fast lookup */
	int dense_y[data->N];
	if (!only_classify)
	{
		memset(dense_y, 0, sizeof(int)*data->N);
		for (int i=0; i<gt_sparsity; i++)
			dense_y[y[i+1]] = 1;
	}

	for(int image_nr=0; image_nr<data->N; image_nr++)
	{
		/* score = <phi(x^n),theta(tag)> */
		double *theta_tag      = theta+tag*data->x_dim;
		double  score          = 0;

		if (sparse_features)
		{
			double *vals = data->features_vals[image_nr];
			int    *idxs = data->features_idxs[image_nr];
			int     nnz  = data->features_nnz[image_nr];

#ifdef USE_MKL
			score = cblas_ddoti(nnz, vals, idxs, theta_tag);
#else
			for (int i=0; i<nnz; i++)
				score += theta_tag[idxs[i]]*vals[i];
#endif

		}
		else
		{
			double *image_features = data->features[image_nr];
#ifdef USE_MKL
			score = cblas_ddot(data->x_dim, theta_tag, 1, image_features, 1);
#else
			for (int i=0; i<data->x_dim; i++)
				score += theta_tag[i]*image_features[i];
#endif
		}
		image_scores[image_nr] = score;

	}

	if (only_classify)
	{
		for(int image_nr=0; image_nr<data->N; image_nr++)
		{
			scores[image_nr].score = image_scores[image_nr];
			scores[image_nr].id = image_nr;
		}
		qsort(scores, data->N, sizeof(score_t), compare_scores);

		for(ybar[0]=0; ybar[0]<data->N; ybar[0]++)
		{
			if ((scores[ybar[0]].score<=0)&&(ybar[0]>0)) break;
			ybar[ybar[0]+1] = scores[ybar[0]].id;
		}
	}
	else
	{
		if (score_to_optimize==MACRO_FBETA)
		{
			double sparsity_score, best_sparsity_score=-1e100;
			int min_sparsity = 0;
			int calls_to_is_sparsity_bound = 0;

			/* Get the indexes of the images which have a ground truth of zero and a positive score.
			 *
			 * For these z_n doesn't depend on the second term, so they must be included
			 * in ybar, as removing them would both:
			 * - reduce the lhs term of eq. (9) (since they have a positive score)
			 * - and decrease the negative term in the rhs of eq. (9) of all images whose ground truth is one.
			 * The argmax (eq. (8)) would, therefore, decrease.
			 *
			 * This gives us an initial lower bound on the sparsity of ybar.
			 */
			for (int image_nr = 0; image_nr<data->N; image_nr++)
				min_sparsity += (dense_y[image_nr]==0)&&(image_scores[image_nr]>0);

			if (min_sparsity<1)
			{
				/* special case: at first iteration, with all weights set to zero, we fix min_sparsity to 1 */
				min_sparsity=1;
			}
			else
			{
				/* Count the number of positive entries in z_n for sparsity=data->N; that is an upper bound
				 * on the number of positive entries we can have in z_n, and therefore also an upper bound
				 * on our lower bound on sparsity (see is_sparsity_bound). */
				int max_num_pos = 0;
				double score, const_factor=-(1+beta_sq)/(beta_sq*gt_sparsity+data->N);

				for(int image_nr=0; image_nr<data->N; image_nr++)
				{
					score = image_scores[image_nr];
					if (dense_y[image_nr]==1) score += const_factor;
					if (score>0) max_num_pos++;
				}

#if 1 /* both methods (below and after the #else) have very similar run times; whatever is chosen the final result (ybar) will be the same */
				/* do a binary search to find a better lower bound on the sparsity */
				int first = min_sparsity;
				int last  = max_num_pos;

				while (first <= last)
				{
					int middle = (first+last)/2;
					if (is_sparsity_bound(middle, gt_sparsity, image_scores, dense_y, beta_sq))
					{
						min_sparsity = middle;
						first = middle+1;  /* repeat search in top half */
					}
					else
					{
						last = middle-1; /* repeat search in bottom half */
					}
					calls_to_is_sparsity_bound++;
				}
#else
				/* find a better lower bound on the sparsity using algorithm 3 */
				int k = min_sparsity;
				do
				{
					int pos = is_sparsity_bound(k, gt_sparsity, image_scores, dense_y, beta_sq);
					calls_to_is_sparsity_bound++;

					if (k<pos)
					{
						min_sparsity = pos;
						k = (max_num_pos+min_sparsity)/2;
					}
					
					if (k>pos)
					{
						k = (k+min_sparsity)/2;
					}
				} while (k!=min_sparsity);
#endif
			}
			if (min_sparsity>=max_sparsity) min_sparsity=max_sparsity;

			/* with the new lower bound, do an exhaustive search */
			for (int sparsity=min_sparsity; sparsity<=max_sparsity; sparsity++)
			{
				for(int image_nr=0; image_nr<data->N; image_nr++)
				{
					scores[image_nr].score = image_scores[image_nr];

					/* loss */
					if (dense_y[image_nr]==1)
						scores[image_nr].score += -(1+beta_sq)/(beta_sq*gt_sparsity+sparsity);

					scores[image_nr].id = image_nr;
				}

				qsort(scores, data->N, sizeof(score_t), compare_scores);

				sparsity_score = 0.0;
				for(int i=0; i<sparsity; i++) 
					sparsity_score += scores[i].score;

				if (sparsity_score>best_sparsity_score)
				{
					best_sparsity_score = sparsity_score;
					ybar[0]             = sparsity;
					for(int i=0; i<sparsity; i++) 
						ybar[i+1] = scores[i].id;
				}
			}
			//fprintf(stderr, "%d %d %d %d\n", min_sparsity, ybar[0], data->N, calls_to_is_sparsity_bound);
		}

		if (score_to_optimize==HAMMING_LOSS)
		{
			/* with Hamming loss is easier: the scores don't depend on the number of images selected, so
			 * we just have to compute the scores once and find the images that have positive scores */

			for(int image_nr=0; image_nr<data->N; image_nr++)
			{
				scores[image_nr].score = image_scores[image_nr];

				/* loss */
				/* the reasoning is the following:
				 * - we start with a baseline prediction of all zeros
				 * - we iterate over all images, and compute the effect on the score of changing the prediction to 1
				 * - if the ground truth is zero, we will increase the loss
				 * - if the ground truth is one,  we will decrease the loss
				 */
				if (dense_y[image_nr]==1)
					scores[image_nr].score += -1.0/((double)data->N);
				else
					scores[image_nr].score += +1.0/((double)data->N);

				scores[image_nr].id = image_nr;
			}
			qsort(scores, data->N, sizeof(score_t), compare_scores);

			for(ybar[0]=0; ybar[0]<data->N; ybar[0]++)
			{
				if (scores[ybar[0]].score<0) break;
				ybar[ybar[0]+1] = scores[ybar[0]].id;
			}

		}
	}


	/* sort ybar in ascending order (we assume that in label_loss) */
	qsort(ybar+1, ybar[0], sizeof(int), compare_int);
}

void CGenericLoss::ComputeLossAndGradient(double& loss, TheMatrix& grad)
{
	double *g    = grad.Data();

	loss = 0;
	grad.Zero();

#define MAX_N_CPUS 8

	double loss_c[MAX_N_CPUS];
	double *g_c[MAX_N_CPUS];
	int     dim = data->x_dim*data->vocab_size;
	int     used_cpus = 0;

	for(int cpu=0; cpu<MAX_N_CPUS; cpu++)
	{
		g_c[cpu] = new double[dim];
		memset(g_c[cpu], 0, sizeof(double)*dim);
		loss_c[cpu] = 0;
	}

#pragma omp parallel for
	for (int tag=0; tag<data->vocab_size; tag++)
	{
#ifdef USE_OPENMP
		int cpu = omp_get_thread_num();
#else
		int cpu = 0;
#endif
		if (cpu>=MAX_N_CPUS)
		{
			printf("cpu = %d, aborting\n", cpu);
			exit(-1);
		}
		if (cpu>used_cpus)
			used_cpus = cpu;

		int *ybar = new int[max_sparsity+1];
		int *y    = data->images_with_tag[tag];

		/* find the y that generates the most violated constraint (we'll call it ybar) */
		find_most_violated_constraint(tag, ybar, _model->GetW(), false);

		/* add phi(x^n,ybar) to the gradient */
		for (int i=0; i<ybar[0]; i++)
		{
			int image_nr = ybar[i+1];
			int offset   = tag*data->x_dim;
			double fact  = data->vocab_size;

			if (sparse_features)
			{
				double *vals = data->features_vals[image_nr];
				int    *idxs = data->features_idxs[image_nr];
				int     nnz  = data->features_nnz[image_nr];

				for (int j=0; j<nnz; j++)
					g_c[cpu][offset+idxs[j]] += vals[j]/fact;
			}
			else
			{
				double  *x   = data->features[image_nr];
#ifdef USE_MKL
				cblas_daxpy(data->x_dim, 1.0/fact, x, 1, g_c[cpu]+offset, 1);
#else
				for (int j=0; j<data->x_dim; j++)
					g_c[cpu][offset+j] += x[j]/fact;
#endif
			}
		}

		/* add -phi(x^n,y^n) to the gradient */
		for (int i=0; i<y[0]; i++)
		{
			int image_nr = y[i+1];
			int offset   = tag*data->x_dim;
			double fact  = data->vocab_size;

			if (sparse_features)
			{
				double *vals = data->features_vals[image_nr];
				int    *idxs = data->features_idxs[image_nr];
				int     nnz  = data->features_nnz[image_nr];

				for (int j=0; j<nnz; j++)
					g_c[cpu][offset+idxs[j]] += -vals[j]/fact;
			}
			else
			{
				double  *x   = data->features[image_nr];
#ifdef USE_MKL
				cblas_daxpy(data->x_dim, -1.0/fact, x, 1, g_c[cpu]+offset, 1);
#else
				for (int j=0; j<data->x_dim; j++)
					g_c[cpu][offset+j] += -x[j]/fact;
#endif
			}
		}

		/* add loss(ybar, y^n) to the loss */
		loss_c[cpu] += label_loss(y, ybar);
		delete[] ybar;
	}
	used_cpus++;

	for(int cpu=0; cpu<used_cpus; cpu++)
	{
		loss += loss_c[cpu];
#ifdef USE_MKL
		cblas_daxpy(dim, 1.0, g_c[cpu], 1, g, 1);
#else
		for (int j=0; j<dim; j++)
			g[j] += g_c[cpu][j];
#endif
	}

	for(int cpu=0; cpu<MAX_N_CPUS; cpu++)
		delete[] g_c[cpu];

	loss = loss/data->vocab_size;
	//printf("loss = %f\n", loss);
}

void CGenericLoss::Predict(CModel *model)
{
	printf("In %s!\n", __FUNCTION__);
	Evaluate(model);
}

void CGenericLoss::Evaluate(CModel *model)
{
	int    *ybar      = new int[data->N+1];
	FILE   *f         = NULL;
	double  ll        = 0.0;
	double  precision = 0.0;
	double  recall    = 0.0;

	if (verbosity>1) f = fopen("output.txt", "w");

	for (int tag=0; tag<data->vocab_size; tag++)
	{
		int *y = data->images_with_tag[tag];

		/* find the most likely y (prediction) */
		find_most_violated_constraint(tag, ybar, _model->GetW(), true);

		/* compute loss */
		ll += label_loss(y, ybar, &precision, &recall);

		if (verbosity>1) { for(int j=1; j<=ybar[0]; j++) fprintf(f, "%d ", ybar[j]); fprintf(f, "\n"); }
	}
	ll        /= data->vocab_size;
	precision /= data->vocab_size;
	recall    /= data->vocab_size;
	if (verbosity>1) fclose(f);

	if (score_to_optimize==MACRO_FBETA)
	{
		printf("macro precision: %f\n", precision);
		printf("macro recall   : %f\n", recall);
		printf("beta           : %f\n", sqrt(beta_sq));
		printf("macro Fbeta    : %f\n", 1-ll);
	}

	if (score_to_optimize==HAMMING_LOSS)
		printf("macro Hamming loss: %f\n", ll);
}

